@@ -10,7 +10,7 @@
 -----------------------------------------------------------------------------
 -- |
 -- Module      :  Control.Comonad.Trans.Cofree
--- Copyright   :  (C) 2008-2012 Edward Kmett
+-- Copyright   :  (C) 2008-2013 Edward Kmett
 -- License     :  BSD-style (see the file LICENSE)
 --
 -- Maintainer  :  Edward Kmett <ekmett@gmail.com>
@@ -21,6 +21,7 @@
 ----------------------------------------------------------------------------
 module Control.Comonad.Trans.Cofree
   ( CofreeT(..)
+  , Cofree, cofree, runCofree
   , CofreeF(..)
   , ComonadCofree(..)
   , headF
@@ -37,13 +38,14 @@ import Data.Bifunctor
 import Data.Bifoldable
 import Data.Bitraversable
 import Data.Foldable
+import Data.Functor.Identity
 import Data.Semigroup
 import Data.Traversable
+import Control.Monad (liftM)
+import Control.Monad.Trans
+import Control.Monad.Zip
 import Prelude hiding (id,(.))
-
-#if defined(GHC_TYPEABLE) || __GLASGOW_HASKELL__ >= 707
 import Data.Data
-#endif
 
 infixr 5 :<
 
@@ -83,6 +85,39 @@ instance Traversable f => Bitraversable (CofreeF f) where
 
 -- | This is a cofree comonad of some functor @f@, with a comonad @w@ threaded through it at each level.
 newtype CofreeT f w a = CofreeT { runCofreeT :: w (CofreeF f a (CofreeT f w a)) }
+#if __GLASGOW_HASKELL__ >= 707
+  deriving Typeable
+#endif
+
+-- | The cofree `Comonad` of a functor @f@.
+type Cofree f = CofreeT f Identity
+
+{- |
+Wrap another layer around a cofree comonad value.
+
+@cofree@ is a right inverse of `runCofree`.
+
+@
+runCofree . cofree == id
+@
+-}
+cofree :: CofreeF f a (Cofree f a) -> Cofree f a
+cofree = CofreeT . Identity
+{-# INLINE cofree #-}
+
+
+{- |
+Unpeel the first layer off a cofree comonad value.
+
+@runCofree@ is a right inverse of `cofree`.
+
+@
+cofree . runCofree == id
+@
+-}
+runCofree :: Cofree f a -> CofreeF f a (Cofree f a)
+runCofree = runIdentity . runCofreeT
+{-# INLINE runCofree #-}
 
 instance (Functor f, Functor w) => Functor (CofreeT f w) where
   fmap f = CofreeT . fmap (bimap f (fmap f)) . runCofreeT
@@ -117,11 +152,36 @@ instance Eq (w (CofreeF f a (CofreeT f w a))) => Eq (CofreeT f w a) where
 instance Ord (w (CofreeF f a (CofreeT f w a))) => Ord (CofreeT f w a) where
   compare (CofreeT a) (CofreeT b) = compare a b
 
+instance (Alternative f, Monad w) => Monad (CofreeT f w) where
+  return = CofreeT . return . (:< empty)
+  {-# INLINE return #-}
+  CofreeT cx >>= f = CofreeT $ do
+    a :< m <- cx
+    b :< n <- runCofreeT $ f a
+    return $ b :< (n <|> fmap (>>= f) m)
+
+
+instance (Alternative f, Applicative w) => Applicative (CofreeT f w) where
+  pure = CofreeT . pure . (:< empty)
+  {-# INLINE pure #-}
+  wf <*> wa = CofreeT $ go <$> runCofreeT wf <*> runCofreeT wa where
+    go (f :< t) a = case bimap f (fmap f) a of
+      b :< n -> b :< (n <|> fmap (<*> wa) t)
+  {-# INLINE (<*>) #-}
+
+instance Alternative f => MonadTrans (CofreeT f) where
+  lift = CofreeT . liftM (:< empty)
+
+instance (Alternative f, MonadZip f, MonadZip m) => MonadZip (CofreeT f m) where
+  mzip (CofreeT ma) (CofreeT mb) = CofreeT $ do
+                                     (a :< fa, b :< fb) <- mzip ma mb
+                                     return $ (a, b) :< (uncurry mzip <$> mzip fa fb)
+
 -- | Unfold a @CofreeT@ comonad transformer from a coalgebra and an initial comonad.
 coiterT :: (Functor f, Comonad w) => (w a -> f (w a)) -> w a -> CofreeT f w a
-coiterT psi = CofreeT . (extend $ \w -> extract w :< fmap (coiterT psi) (psi w))
+coiterT psi = CofreeT . extend (\w -> extract w :< fmap (coiterT psi) (psi w))
 
-#if defined(GHC_TYPEABLE) && __GLASGOW_HASKELL__ < 707
+#if __GLASGOW_HASKELL__ < 707
 
 instance Typeable1 f => Typeable2 (CofreeF f) where
   typeOf2 t = mkTyConApp cofreeFTyCon [typeOf1 (f t)] where
@@ -146,6 +206,10 @@ cofreeFTyCon = mkTyCon3 "free" "Control.Comonad.Trans.Cofree" "CofreeF"
 {-# NOINLINE cofreeTTyCon #-}
 {-# NOINLINE cofreeFTyCon #-}
 
+#else
+#define Typeable1 Typeable
+#endif
+
 instance
   ( Typeable1 f, Typeable a, Typeable b
   , Data a, Data (f b), Data b
@@ -182,7 +246,6 @@ cofreeFDataType = mkDataType "Control.Comonad.Trans.Cofree.CofreeF" [cofreeFCons
 cofreeTDataType = mkDataType "Control.Comonad.Trans.Cofree.CofreeT" [cofreeTConstr]
 {-# NOINLINE cofreeFDataType #-}
 {-# NOINLINE cofreeTDataType #-}
-#endif
 
 -- lowerF :: (Functor f, Comonad w) => CofreeT f w a -> f a
 -- lowerF = fmap extract . unwrap
