@@ -1,7 +1,11 @@
+{-# LANGUAGE CPP #-}
 {-# LANGUAGE Rank2Types #-}
 {-# LANGUAGE FlexibleInstances #-}
 {-# LANGUAGE UndecidableInstances #-}
 {-# LANGUAGE MultiParamTypeClasses #-}
+#ifndef MIN_VERSION_base
+#define MIN_VERSION_base(x,y,z) 1
+#endif
 -----------------------------------------------------------------------------
 -- |
 -- Module      :  Control.Monad.Free.Church
@@ -14,10 +18,38 @@
 --
 -- \"Free Monads for Less\"
 --
--- This is based on the \"Free Monads for Less\" series of articles:
+-- The most straightforward way of implementing free monads is as a recursive
+-- datatype that allows for arbitrarily deep nesting of the base functor. This is
+-- akin to a tree, with the leaves containing the values, and the nodes being a
+-- level of 'Functor' over subtrees.
 --
--- <http://comonad.com/reader/2011/free-monads-for-less/>
--- <http://comonad.com/reader/2011/free-monads-for-less-2/>
+-- For each time that the `fmap` or `>>=` operations is used, the old tree is
+-- traversed up to the leaves, a new set of nodes is allocated, and
+-- the old ones are garbage collected. Even if the Haskell runtime
+-- optimizes some of the overhead through laziness and generational garbage
+-- collection, the asymptotic runtime is still quadratic.
+--
+-- On the other hand, if the Church encoding is used, the tree only needs to be
+-- constructed once, because:
+--
+-- * All uses of `fmap` are collapsed into a single one, so that the values on the
+--   _leaves_ are transformed in one pass.
+--
+--   prop> fmap f . fmap g == fmap (f . g)
+--
+-- * All uses of `>>=` are right associated, so that every new subtree created
+--   is final.
+--
+--   prop> (m >>= f) >>= g == m >>= (\x -> f x >>= g)
+--
+-- Asymptotically, the Church encoding supports the monadic operations more
+-- efficiently than the naïve 'Free'.
+--
+-- This is based on the \"Free Monads for Less\" series of articles by Edward Kmett:
+--
+-- * <http://comonad.com/reader/2011/free-monads-for-less/   Free monads for less — Part 1>
+--
+-- * <http://comonad.com/reader/2011/free-monads-for-less-2/ Free monads for less — Part 2>
 ----------------------------------------------------------------------------
 module Control.Monad.Free.Church
   ( F(..)
@@ -26,19 +58,23 @@ module Control.Monad.Free.Church
   , iterM
   , toF
   , retract
+  , hoistF
   , MonadFree(..)
   , liftF
   ) where
 
 import Control.Applicative
 import Control.Monad as Monad
+import Control.Monad.Fix
 import Control.Monad.Free hiding (retract, iterM)
 import Control.Monad.Reader.Class
 import Control.Monad.Writer.Class
 import Control.Monad.Cont.Class
 import Control.Monad.Trans.Class
 import Control.Monad.State.Class
+import Data.Foldable
 import Data.Functor.Bind
+import Prelude hiding (foldr)
 
 -- | The Church-encoded free monad for a functor @f@.
 --
@@ -59,7 +95,7 @@ instance Apply (F f) where
 
 instance Applicative (F f) where
   pure a = F (\kp _ -> kp a)
-  F f <*> F g = F (\kp kf -> f (\a -> g (\b -> kp (a b)) kf) kf)
+  F f <*> F g = F (\kp kf -> f (\a -> g (kp . a) kf) kf)
 
 instance Alternative f => Alternative (F f) where
   empty = F (\_ kf -> kf empty)
@@ -72,6 +108,21 @@ instance Monad (F f) where
   return a = F (\kp _ -> kp a)
   F m >>= f = F (\kp kf -> m (\a -> runF (f a) kp kf) kf)
 
+instance MonadFix (F f) where
+  mfix f = a where
+    a = f (impure a)
+    impure (F x) = x id (error "MonadFix (F f): wrap")
+
+instance (Foldable f, Functor f) => Foldable (F f) where
+    foldr f r xs = runF xs f (foldr (.) id) r
+    {-# INLINE foldr #-}
+
+#if MIN_VERSION_base(4,6,0)
+    foldl' f z xs = runF xs (flip f) (foldr (!>>>) id) z
+      where (!>>>) h g = \r -> g $! h r
+    {-# INLINE foldl' #-}
+#endif
+
 instance MonadPlus f => MonadPlus (F f) where
   mzero = F (\_ kf -> kf mzero)
   F f `mplus` F g = F (\kp kf -> kf (return (f kp kf) `mplus` return (g kp kf)))
@@ -109,6 +160,10 @@ retract :: Monad m => F m a -> m a
 retract (F m) = m return Monad.join
 {-# INLINE retract #-}
 
+-- | Lift a natural transformation from @f@ to @g@ into a natural transformation from @F f@ to @F g@.
+hoistF :: (forall x. f x -> g x) -> F f a -> F g a
+hoistF t (F m) = F (\p f -> m p (f . t))
+
 -- | Convert to another free monad representation.
 fromF :: MonadFree f m => F f a -> m a
 fromF (F m) = m return wrap
@@ -124,12 +179,11 @@ toF xs = F (\kp kf -> go kp kf xs) where
 --
 -- This is based on the \"Free Monads for Less\" series of articles by Edward Kmett:
 --
--- <http://comonad.com/reader/2011/free-monads-for-less/>
--- <http://comonad.com/reader/2011/free-monads-for-less-2/>
+-- * <http://comonad.com/reader/2011/free-monads-for-less/   Free monads for less — Part 1>
 --
--- and \"Asymptotic Improvement of Computations over Free Monads\" by Janis Voightländer:
+-- * <http://comonad.com/reader/2011/free-monads-for-less-2/ Free monads for less — Part 2>
 --
--- <http://www.iai.uni-bonn.de/~jv/mpc08.pdf>
+-- and <http://www.iai.uni-bonn.de/~jv/mpc08.pdf \"Asymptotic Improvement of Computations over Free Monads\"> by Janis Voightländer.
 improve :: Functor f => (forall m. MonadFree f m => m a) -> Free f a
 improve m = fromF m
 {-# INLINE improve #-}
